from redis.asyncio import Redis, BlockingConnectionPool

from aioscpy.queue import BaseQueue


class PriorityQueue(BaseQueue):
    def __init__(self, server, spider, key=None, serializer="pickle"):
        super().__init__(server, spider, key)
        self.serializer = self.__compat__[serializer]

    async def qsize(self) -> int:
        return await self.server.zcard(self.key)

    async def push(self, request):
        data = self._encode_request(request)
        score = -request.get('priority', 1)
        await self.server.zadd(self.key, {data: score})

    async def mpush(self, requests: list):
        async with self.server.pipeline() as pipe:
            for request in requests:
                data = self._encode_request(request)
                score = -request.get('priority', 1)
                pipe.zadd(self.key, {data: score})
            await pipe.execute()

    async def pop(self, timeout: int = 0, count: int = 0):
        async with self.server.pipeline(transaction=True) as pipe:
            results, _ = await (
                pipe.zrange(self.key, 0, count)
                    .zremrangebyrank(self.key, 0, count)
                    .execute()
            )
        _results = []
        for result in results:
            _results.append(self._decode_request(result))
        return _results


class AsyncRedis:
    __redis_instance = None

    def __init__(self, *args, **kwargs):
        self.args = args
        if not kwargs:
            self.kwargs = self.validator(kwargs)
        self.kwargs = kwargs

    @staticmethod
    def validator(params: dict) -> dict:
        params.setdefault('host', '127.0.0.1')
        params.setdefault('port', 6379)
        params.setdefault('db', 1)
        params.setdefault('password', 'admin')
        return params

    @property
    async def get_redis_pool(self) -> Redis:
        if not self.__redis_instance:
            url = self.kwargs.pop('url', None)
            if url:
                connection_pool = BlockingConnectionPool.from_url(url, **self.kwargs)
            else:
                connection_pool = BlockingConnectionPool(**self.kwargs)
            self.__redis_instance = Redis(connection_pool=connection_pool)
        return self.__redis_instance

    async def close(self):
        if self.__redis_instance:
            await self.__redis_instance.close()


async def aio_priority_queue(key: str, redis_tcp, spider) -> PriorityQueue:
    """
    # unit test example
    async def run():
        REDIS_TCP = {
                    "host": "172.16.7.172",
                    "port": 6379,
                    "password": "123456",
                    "db": 15
                }
        queue = await aio_priority_queue('message:queue', REDIS_TCP)
        # await queue.push({"url": "https://www.baidu.com/?kw=1", "task_id": '123'})
        print(await queue.pop())


    if __name__ == "__main__":
        import asyncio
        asyncio.run(run())
    """

    if isinstance(redis_tcp, str):
        redis_tcp = {'url': redis_tcp}
    server = await AsyncRedis(**redis_tcp).get_redis_pool
    return PriorityQueue(server=server, spider=spider, key=key, serializer='json')


spider_aio_priority_queue = aio_priority_queue



import asyncio
import ssl
import httpx

from anti_header import Headers
from anti_useragent.utils.cipers import generate_cipher


class HttpxDownloadHandler(object):

    def __init__(self, settings, crawler):
        self.settings = settings
        self.crawler = crawler
        self.context = ssl.create_default_context()

    @classmethod
    def from_settings(cls, settings, crawler):
        return cls(settings, crawler)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings, crawler)

    async def download_request(self, request, spider):
        headers = request.headers
        if isinstance(headers, Headers):
            headers = headers.to_unicode_dict()
        httpx_client_session = {}

        if request.meta.get('TLS_CIPHERS') or self.settings.get('TLS_CIPHERS'):
            self.context.set_ciphers(generate_cipher())
            httpx_client_session['verify'] = self.context

        if request.meta.get("proxy"):
            httpx_client_session['proxies'] = request.meta["proxy"]
            self.logger.debug(f"use {request.meta['proxy']} crawling: {request.url}")

        session_kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'headers': headers,
            'follow_redirects': True
        }
        if isinstance(request.body, dict):
            session_kwargs['json'] = request.body or None
        else:
            session_kwargs['data'] = request.body or None
        
        async with httpx.AsyncClient(**httpx_client_session) as session:
            response = await session.request(request.method, request.url, **session_kwargs)
            content = response.read()

        return self.di.get("response")(
            str(response.url),
            status=response.status_code,
            headers=response.headers,
            body=content,
            cookies=response.cookies,
            _response=response)

    async def close(self):
        await asyncio.sleep(0.1)



from aioscpy.core.scheduler import Scheduler
from aioscpy.queue.redis import aio_priority_queue


class RedisScheduler(Scheduler):

    @classmethod
    def from_crawler(cls, crawler):
        redis_tcp = crawler.settings.get('REDIS_URI') or \
                    crawler.settings.get('REDIS_TCP')
        queue_key = crawler.settings.get('QUEUE_KEY') % {'spider': crawler.spider.name}
        return cls(_queue_df=aio_priority_queue(queue_key, redis_tcp, crawler.spider), spider=crawler.spider, stats=crawler.stats)

    async def has_pending_requests(self):
        return await self.queue.qsize() > 0



from urllib.parse import urljoin, urlencode

from aioscpy.http.request import Request
from aioscpy.utils.tools import to_bytes, is_listlike


class JsonRequest(Request):
    valid_form_methods = ['GET', 'POST']

    def __init__(self, *args, **kwargs):
        formdata = kwargs.pop('formdata', None)
        if formdata and kwargs.get('method') is None:
            kwargs['method'] = 'POST'

        super(JsonRequest, self).__init__(*args, **kwargs)

        if formdata:
            if self.method == 'POST':
                self.headers.setdefault(
                    b'Content-Type', [b'application/json'])
                self._set_body(formdata)




from w3lib.url import safe_url_string


class Request(object):

    def __init__(self, url,
                 callback=None,
                 method='GET',
                 headers=None,
                 body=None,
                 cookies=None,
                 meta=None,
                 encoding='utf-8',
                 priority=0,
                 dont_filter=False,
                 errback=None, flags=None, cb_kwargs=None):
        self._encoding = encoding  # this one has to be set first
        self.method = str(method).upper()
        self._set_url(url)
        self._set_body(body)
        assert isinstance(priority, int), "Request priority not an integer: %r" % priority
        self.priority = priority

        if callback is not None and not callable(callback):
            raise TypeError('callback must be a callable, got %s' %
                            type(callback).__name__)
        if errback is not None and not callable(errback):
            raise TypeError('errback must be a callable, got %s' %
                            type(errback).__name__)
        self.callback = callback
        self.errback = errback

        self.cookies = cookies or {}
        self.headers = headers or {}
        self.dont_filter = dont_filter

        self._meta = dict(meta) if meta else None
        self._cb_kwargs = dict(cb_kwargs) if cb_kwargs else None
        self.flags = [] if flags is None else list(flags)

    @property
    def cb_kwargs(self):
        if self._cb_kwargs is None:
            self._cb_kwargs = {}
        return self._cb_kwargs

    @property
    def meta(self):
        if self._meta is None:
            self._meta = {}
        return self._meta

    def get(self, key, default):
        return self.meta.get(key, default)

    def _get_url(self):
        return self._url

    def _set_url(self, url):
        if not isinstance(url, str):
            raise TypeError(
                'Request url must be str or unicode, got %s:' % type(url).__name__)

        s = safe_url_string(url, self.encoding)
        self._url = s

        if ('://' not in self._url) and (not self._url.startswith('data:')):
            raise ValueError('Missing scheme in request url: %s' % self._url)

    url = property(_get_url, _set_url)

    def _get_body(self):
        return self._body

    def _set_body(self, body):
        if body is None:
            self._body = b''
        else:
            self._body = body

    body = property(_get_body, _set_body)

    @property
    def encoding(self):
        return self._encoding

    def __str__(self):
        return "<%s %s>" % (self.method, self.url)

    __repr__ = __str__

    def copy(self):
        """Return a copy of this Request"""
        return self.replace()

    def replace(self, *args, **kwargs):
        """Create a new Request with the same attributes except for those
        given new values.
        """
        for x in ['url', 'method', 'headers', 'body', 'cookies', 'meta', 'flags',
                  'encoding', 'priority', 'dont_filter', 'callback', 'errback', 'cb_kwargs']:
            kwargs.setdefault(x, getattr(self, x))
        cls = kwargs.pop('cls', self.__class__)
        return cls(*args, **kwargs)



import asyncio

from importlib import import_module
from pkgutil import iter_modules

from aioscpy.settings import Settings
from aioscpy.utils.tools import singleton, get_project_settings


@singleton
class CSlot:

    def __init__(self):
        self._object_slot_cls = {}

    def get(self, sets: str, default=None) -> object:
        return self._object_slot_cls.get(sets, default)

    def set(self, sets: str, obj: object):
        self._object_slot_cls.__setitem__(sets, obj)

    def empty(self):
        return not bool(len(self._object_slot_cls))


class Slot:

    def __init__(self, settings, crawler):
        self._objects_slot = {
            'settings': settings,
            'crawler': crawler
        }
        self._modules_slot = []
        self._close = None
        self.live_beat = None

    @property
    def is_live(self):
        return bool(self._close)

    def get(self, sets: str, default=None) -> object:
        return self._objects_slot.get(sets, default)

    def set(self, sets: str, obj: object):
        self._objects_slot.__setitem__(sets, obj)

    def clear(self):
        del self._objects_slot
        self._modules_slot = []
        self._close = True

    def close(self):
        if self.live_beat:
            self.live_beat.cancel()


class DependencyInjection(object):
    def __init__(self, settings: Settings = None, crawler=None):
        if not settings:
            settings = Settings()
        self.settings = settings
        self.crawler = crawler
        self.slot = Slot(settings, crawler)

    @classmethod
    def from_settings(cls, settings, crawler):
        return cls(settings, crawler)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings, crawler)

    def load(self, key: str):
        return self.slot.get(key)

    @staticmethod
    def load_all_spider(dirname):
        _class_objects = {}
    
        def load_all_spider_inner(dirname):
            for importer, package_name, ispkg in iter_modules([dirname]):
                if ispkg:
                    load_all_spider_inner(dirname + '/' + package_name)
                else:
                    module = importer.find_module(package_name)
                    module = module.load_module(package_name)
                    for cls_name in module.__dir__():
                        if cls_name == "__spiders__":
                            class_object = getattr(module, cls_name)
                            for co in class_object:
                                _class_objects[co.name] = co
                        if not cls_name.startswith('__'):
                            class_object = getattr(module, cls_name)
                            if hasattr(class_object, "name") and getattr(class_object, "name"):
                                _class_objects[class_object.name] = class_object


        load_all_spider_inner(dirname)
        return _class_objects

    @staticmethod
    def load_object(path: str):
        try:
            dot = path.rindex('.')
        except ValueError:
            raise ValueError("Error loading object '%s': not a full path" % path)

        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)

        try:
            obj = getattr(mod, name)
        except AttributeError:
            raise NameError("Module '%s' doesn't define any object named '%s'" % (module, name))
        else:
            return obj

    def load_object_slot(self, key: str, path: str, cls=None):
        obj = self.load_object(path)
        if cls is None:
            obj = self.create_instance(obj, self.settings, self.crawler)
            self.slot.set(key, obj)
        else:
            self.c_slot.set(key, obj)
        return obj

    def walk_modules(self, path: str):
        mods = []
        if hasattr(self, "slot"):
            mods = self.slot._modules_slot
        mod = import_module(path)
        mods.append(mod)
        if hasattr(mod, '__path__'):
            for _, subpath, ispkg in iter_modules(mod.__path__):
                fullpath = path + '.' + subpath
                if ispkg:
                    mods += self.walk_modules(fullpath)
                else:
                    submod = import_module(fullpath)
                    mods.append(submod)
        return mods

    def create_instance(self, objcls, settings, crawler, *args, **kwargs):
        if settings is None:
            if crawler is None:
                raise ValueError("Specify at least one of settings and crawler.")
            settings = crawler.settings
        if not (type(objcls) == "function"):
            objcls = call_grace_instance(objcls, only_instance=True)
        if crawler and hasattr(objcls, 'from_crawler'):
            return objcls.from_crawler(crawler, *args, **kwargs)
        elif hasattr(objcls, 'from_settings'):
            return objcls.from_settings(settings, *args, **kwargs)
        else:
            return objcls(*args, **kwargs)

    async def inject_runner(self):
        if any([not self.settings.get('DI_CONFIG'), not self.settings.get('DI_CONFIG_CLS')]):
            raise KeyError('Settings DI_CONFIG/DI_CONFIG_CLS not be None')
        for key, value in self.settings['DI_CONFIG'].items():
            self.load_object_slot(key, value)
        # self.slot.live_beat = asyncio.create_task(self.live_beat())

    async def live_beat(self):
        while 1:
            if not self.slot.is_live:
                await asyncio.sleep(20)
                break
            asyncio.create_task(self.inject_runner())


class DependencyInjectionCls(DependencyInjection):

    def __init__(self):
        self.c_slot = CSlot()
        self.settings = get_project_settings()

    def inject(self):
        if self.c_slot.empty():
            for key, value in self.settings['DI_CONFIG_CLS'].items():
                self.load_object_slot(key, value, cls=True)
        return self.c_slot


_create_dependency = DependencyInjectionCls()
load_object = _create_dependency.load_object
walk_modules = _create_dependency.walk_modules
settings_ins = _create_dependency.settings


class object_ref(type):
    def __init__(msc, *args, **kwargs):
        msc.di = _create_dependency.inject()
        msc.logger = msc.di.get("log").logger
        super().__init__(*args, **kwargs)


def call_grace_instance(obj, *args, only_instance=None, **kwargs):

    if isinstance(obj, str):
        obj = load_object(settings_ins['DI_CREATE_CLS'].get(obj))

    class Inner(obj, metaclass=object_ref):
        pass
    if only_instance is None:
        return Inner(*args, **kwargs)
    else:
        return Inner



from urllib.parse import urljoin, urlencode

from aioscpy.http.request import Request
from aioscpy.utils.tools import to_bytes, is_listlike


class FormRequest(Request):
    valid_form_methods = ['GET', 'POST']

    def __init__(self, *args, **kwargs):
        formdata = kwargs.pop('formdata', None)
        if formdata and kwargs.get('method') is None:
            kwargs['method'] = 'POST'

        super(FormRequest, self).__init__(*args, **kwargs)

        if formdata:
            items = formdata.items() if isinstance(formdata, dict) else formdata
            querystr = _urlencode(items, self.encoding)
            if self.method == 'POST':
                self.headers.setdefault(
                    b'Content-Type', [b'application/x-www-form-urlencoded'])
                self._set_body(to_bytes(querystr, self.encoding))
            else:
                self._set_url(
                    self.url + ('&' if '?' in self.url else '?') + querystr)


def _urlencode(seq, enc):
    values = [(to_bytes(k, enc), to_bytes(v, enc))
              for k, vs in seq
              for v in (vs if is_listlike(vs) else [vs])]
    return urlencode(values, doseq=1)



from redis import ConnectionPool, StrictRedis

from aioscpy.queue import BaseQueue


class PriorityQueue(BaseQueue):
    def qsize(self) -> int:
        """Return the length of the queue"""
        return self.server.zcard(self.key)

    def push(self, request: dict):
        data = self._encode_request(request)
        score = -request.get('priority', 1)
        self.server.zadd(self.key, {data: score})

    def pop(self, timeout: int = 0) -> dict:
        pipe = self.server.pipeline()
        pipe.multi()
        pipe.zrange(self.key, 0, 0).zremrangebyrank(self.key, 0, 0)
        results, count = pipe.execute()
        if results:
            return self._decode_request(results[0])


class Redis:

    __redis_instance = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = self.validator(kwargs)

    @staticmethod
    def validator(params: dict) -> dict:
        params.setdefault('host', '127.0.0.1')
        params.setdefault('port', 6379)
        params.setdefault('db', 1)
        params.setdefault('password', 'admin')
        return params

    @property
    def format_url(self) -> str:
        """REDIS_URL = 'redis://:123456@172.16.8.147:6379/1'"""
        _format_url = f"redis://:{self.kwargs['password']}@{self.kwargs['host']}:{self.kwargs['port']}/{self.kwargs['db']}"\
            if not self.kwargs.get('redis_url') else self.kwargs['redis_url']
        return _format_url

    @property
    def get_redis_pool(self) -> StrictRedis:
        if not self.__redis_instance:
            pool = ConnectionPool(**self.kwargs)
            self.__redis_instance = StrictRedis(connection_pool=pool)

        return self.__redis_instance

    def close(self):
        if self.__redis_instance:
            self.__redis_instance.close()


def priority_queue(key: str, redis_tcp: dict) -> PriorityQueue:
    """
    def run():
        queue = redis_client('message:queue')
        # queue.push({"url": "https://www.baidu.com/?kw=1", "task_id": '123'})
        print(queue.pop())

    run()
    """
    server = Redis(**redis_tcp).get_redis_pool
    return PriorityQueue(server=server, key=key)


spider_priority_queue = priority_queue



BOT_NAME = "aioscpy"

CONCURRENT_REQUESTS = 16
CONCURRENT_REQUESTS_PER_DOMAIN = 8
CONCURRENT_REQUESTS_PER_IP = 0
RANDOMIZE_DOWNLOAD_DELAY = True

CONCURRENT_ITEMS = 16

DOWNLOAD_DELAY = 0
DOWNLOAD_TIMEOUT = 20
# DOWNLOAD_HANDLER = "aioscpy.core.downloader.handlers.aiohttp.AioHttpDownloadHandler"
DOWNLOAD_HANDLER = "aioscpy.core.downloader.handlers.httpx.HttpxDownloadHandler"
# SCHEDULER = "aioscpy.core.scheduler.redis.RedisScheduler"
SCHEDULER = "aioscpy.core.scheduler.memory.MemoryScheduler"
REQUESTS_SESSION_STATS = False


SPIDER_IDLE = False

# LOG CONFIG
LOG_LEVEL = "DEBUG"
LOG_FILE = False
LOG_FILENAME = f"{BOT_NAME}.log"
LOG_ENCODING = "utf-8"
LOG_ROTATION = "1 week"
LOG_RETENTION = "30 days"

DI_CONFIG = {
    "scheduler": f"{SCHEDULER}",
    "log_formatter": "aioscpy.logformatter.LogFormatter",
    "extension": "aioscpy.middleware.ExtensionManager",

}
DI_CONFIG_CLS = {
    "request": "aioscpy.http.Request",
    "response": "aioscpy.http.TextResponse",
    "form_request": "aioscpy.http.FormRequest",
    "json_request": "aioscpy.http.JsonRequest",
    "logger": "aioscpy.utils.log.logger",
    "log": "aioscpy.utils.log",
    "exceptions": "aioscpy.exceptions",
    "tools": "aioscpy.utils.tools",
    'downloader_middleware': 'aioscpy.middleware.DownloaderMiddlewareManager',
    "item_processor": "aioscpy.middleware.ItemPipelineManager",
}
DI_CREATE_CLS = {
    'crawler': 'aioscpy.crawler.Crawler',
    'crawler_process': 'aioscpy.crawler.CrawlerProcess',
    'engine': 'aioscpy.core.engine.ExecutionEngine',
    'spider': 'aioscpy.spider.Spider',
    'downloader_handler': f'{DOWNLOAD_HANDLER}',
    'stats': 'aioscpy.libs.statscollectors.MemoryStatsCollector',
    'scraper': 'aioscpy.core.scraper.Scraper',
    "downloader": "aioscpy.core.downloader.Downloader",
}

# message config
# RABBITMQ_TCP = {
#     "host": "172.16.8.147",
#     # "port": 5672,
#     # "username": "admin",
#     # "password": "admin",
#     # "key": "message:queue",
#     # "max_priority": 100
# }
QUEUE_KEY = '%(spider)s:requests'

# REDIS_TCP = {
#     "host": "172.16.7.172",
#     "port": 6379,
#     "password": "123456",
#     "db": 15
# }
# REDIS_URI = "redis://:123456@172.16.7.172:6379/1"


EXTENSIONS_BASE = {
    'aioscpy.libs.extensions.corestats.CoreStats': 0,
    'aioscpy.libs.extensions.logstats.LogStats': 0,

}

DOWNLOADER_MIDDLEWARES_BASE = {
    # Engine side
    'aioscpy.libs.downloadermiddlewares.stats.DownloaderStats': 850,
    # Downloader side
}
DOWNLOADER_STATS = True

LOGSTATS_INTERVAL = 60.0
STATS_CLASS = 'aioscpy.libs.statscollectors.MemoryStatsCollector'
STATS_DUMP = True
SCRAPER_SLOT_MAX_ACTIVE_SIZE = 5000000

TLS_CIPHERS = False




import asyncio

from aioscpy.spider import Spider
from anti_header import Header
from pprint import pprint, pformat


class BaiduSpider(Spider):
    name = 'baidu'
    custom_settings = {
        "SPIDER_IDLE": False,
        'TLS_CIPHERS': True,
    }
    start_urls = [f'https://www.baidu.com/?a{i}' for i in range(10)]

    async def process_request(self, request):
        request.headers = Header(url=request.url, platform='windows', connection=True).random
        return request

    async def process_response(self, request, response):
        return response

    async def process_exception(self, request, exc):
        raise exc

    async def parse(self, response):
        item = {
            'hot': '\n'.join(response.xpath('//span[@class="title-content-title"]/text()').extract()),
        }
        yield item

    async def process_item(self, item):
        pass
        # self.logger.info("{item}", **{'item': pformat(item)})


if __name__ == '__main__':
    baidu = BaiduSpider()
    baidu.start()



import ujson
import json
import asyncio
import weakref
import sys
import os
import pickle
import warnings

from configparser import ConfigParser
from typing import Dict, Iterable, Optional, Tuple, Union
from functools import wraps
from asyncio import events
from types import CoroutineType, GeneratorType, AsyncGeneratorType

from aioscpy.settings import Settings
from aioscpy.exceptions import AioscpyDeprecationWarning


async def call_create_task(f, *args, **kwargs):
    try:
        if events._get_running_loop():
            await call_helper(asyncio.create_task, f(*args, **kwargs))
    except:
        pass


async def task_await(cls, *args):
    while not all([getattr(cls, arg, None) for arg in args]):
        await asyncio.sleep(0.5)


def install_event_loop_tips():
    if sys.version_info[:2] == (3, 7):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    if sys.platform.startswith('win'):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    else:
        try:
            import uvloop
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        except:
            pass
    setattr(asyncio.sslproto._SSLProtocolTransport, "_start_tls_compatible", True)


def referer_str(request) -> Optional[str]:
    """ Return Referer HTTP header suitable for logging. """
    referrer = request.headers.get('Referer')
    if referrer is None:
        return referrer
    return to_unicode(referrer, errors='replace')


def method_is_overridden(subclass, base_class, method_name):
    """
    Return True if a method named ``method_name`` of a ``base_class``
    is overridden in a ``subclass``.

    >>> class Base(object):
    ...     def foo(self):
    ...         pass
    >>> class Sub1(Base):
    ...     pass
    >>> class Sub2(Base):
    ...     def foo(self):
    ...         pass
    >>> class Sub3(Sub1):
    ...     def foo(self):
    ...         pass
    >>> class Sub4(Sub2):
    ...     pass
    >>> method_is_overridden(Sub1, Base, 'foo')
    False
    >>> method_is_overridden(Sub2, Base, 'foo')
    True
    >>> method_is_overridden(Sub3, Base, 'foo')
    True
    >>> method_is_overridden(Sub4, Base, 'foo')
    True
    """
    base_method = getattr(base_class, method_name)
    sub_method = getattr(subclass, method_name)
    return base_method.__code__ is not sub_method.__code__


def memoizemethod_noargs(method):
    """Decorator to cache the result of a method (without arguments) using a
    weak reference to its object
    """
    cache = weakref.WeakKeyDictionary()

    @wraps(method)
    def new_method(self, *args, **kwargs):
        if self not in cache:
            cache[self] = method(self, *args, **kwargs)
        return cache[self]
    return new_method


def is_listlike(x):
    """
    >>> is_listlike("foo")
    False
    >>> is_listlike(5)
    False
    >>> is_listlike(b"foo")
    False
    >>> is_listlike([b"foo"])
    True
    >>> is_listlike((b"foo",))
    True
    >>> is_listlike({})
    True
    >>> is_listlike(set())
    True
    >>> is_listlike((x for x in range(3)))
    True
    >>> is_listlike(range(5))
    True
    """
    return hasattr(x, "__iter__") and not isinstance(x, (str, bytes))


def unique(list_, key=lambda x: x):
    """efficient function to uniquify a list preserving item order"""
    seen = set()
    result = []
    for item in list_:
        seenkey = key(item)
        if seenkey in seen:
            continue
        seen.add(seenkey)
        result.append(item)
    return result


def to_unicode(text, encoding=None, errors='strict'):
    """Return the unicode representation of a bytes object ``text``. If
    ``text`` is already an unicode object, return it as-is."""
    if isinstance(text, str):
        return text
    if not isinstance(text, (bytes, str)):
        raise TypeError('to_unicode must receive a bytes or str '
                        'object, got %s' % type(text).__name__)
    if encoding is None:
        encoding = 'utf-8'
    return text.decode(encoding, errors)


def to_bytes(text, encoding=None, errors='strict'):
    """Return the binary representation of ``text``. If ``text``
    is already a bytes object, return it as-is."""
    if isinstance(text, bytes):
        return text
    if not isinstance(text, str):
        raise TypeError('to_bytes must receive a str or bytes '
                        'object, got %s' % type(text).__name__)
    if encoding is None:
        encoding = 'utf-8'
    return text.encode(encoding, errors)


async def call_helper(f, *args, **kwargs):
    if asyncio.iscoroutinefunction(f):
        return await f(*args, **kwargs)
    return f(*args, **kwargs)


async def async_generator_wrapper(wrapped):
    """将传入的对象变成AsyncGeneratorType类型"""
    if isinstance(wrapped, AsyncGeneratorType):
        return wrapped

    elif isinstance(wrapped, CoroutineType):
        async def anonymous(c):
            yield await c
        return anonymous(wrapped)

    elif isinstance(wrapped, GeneratorType):
        async def anonymous(c):
            for r in c:
                yield r
        return anonymous(wrapped)

    else:
        async def anonymous(c):
            yield c
        return anonymous(wrapped)


def closest_aioscpy_cfg(path='.', prevpath=None):
    """Return the path to the closest aioscpy.cfg file by traversing the current
    directory and its parents
    """
    if path == prevpath:
        return ''
    path = os.path.abspath(path)
    cfgfile = os.path.join(path, 'aioscpy.cfg')
    if os.path.exists(cfgfile):
        return cfgfile
    return closest_aioscpy_cfg(os.path.dirname(path), path)


def get_sources(use_closest=True):
    xdg_config_home = os.environ.get('XDG_CONFIG_HOME') or os.path.expanduser('~/.config')
    sources = [
        '/etc/aioscpy.cfg',
        r'c:\aioscpy\aioscpy.cfg',
        xdg_config_home + '/aioscpy.cfg',
        os.path.expanduser('~/.aioscpy.cfg'),
    ]
    if use_closest:
        sources.append(closest_aioscpy_cfg())
    return sources


def get_config(use_closest=True):
    """Get Aioscpy config file as a ConfigParser"""
    sources = get_sources(use_closest)
    cfg = ConfigParser()
    cfg.read(sources)
    return cfg


def init_env(project='default', set_syspath=True):
    """Initialize environment to use command-line tool from inside a project
    dir. This sets the Aioscpy settings module and modifies the Python path to
    be able to locate the project module.
    """
    cfg = get_config()
    if cfg.has_option('settings', project):
        os.environ['SETTINGS_MODULE'] = cfg.get('settings', project)
    if cfg.has_option('package_env', "path"):
        path = cfg.get('package_env', 'path')
        sys.path.append(path)
    closest = closest_aioscpy_cfg()
    if closest:
        projdir = os.path.dirname(closest)
        if set_syspath and projdir not in sys.path:
            sys.path.append(projdir)


def get_project_settings():

    ENVVAR = 'SETTINGS_MODULE'

    if ENVVAR not in os.environ:
        project = os.environ.get('AIOSCPY_PROJECT', 'default')
        init_env(project)
    settings = Settings()
    settings_module_path = os.environ.get(ENVVAR)
    if settings_module_path:
        settings.setmodule(settings_module_path, priority='project')

    aioscpy_envvars = {k[8:]: v for k, v in os.environ.items() if
                      k.startswith('AIOSCPY_')}
    valid_envvars = {
        'CHECK',
        'PROJECT',
        'PYTHON_SHELL',
        'SETTINGS_MODULE',
    }
    setting_envvars = {k for k in aioscpy_envvars if k not in valid_envvars}
    if setting_envvars:
        setting_envvar_list = ', '.join(sorted(setting_envvars))
        warnings.warn(
            'Use of environment variables prefixed with AIOSCPY_ to override '
            'settings is deprecated. The following environment variables are '
            f'currently defined: {setting_envvar_list}',
            AioscpyDeprecationWarning
        )
    settings.setdict(aioscpy_envvars, priority='project')
    settings['DI_CONFIG']['scheduler'] = settings['SCHEDULER']
    settings['DI_CREATE_CLS']['downloader_handler'] = settings['DOWNLOAD_HANDLER']
    return settings


def singleton(cls):
    _instance = {}

    def _singleton(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]

    return _singleton


def exec_js_func(js_file_path, func_name, func_params=None, cwd_path=None, cmd_path='node'):
    """
    调用node执行js文件
    :param js_file_path: js文件路径
    :param func_name: 要调用的js文件内的函数名称
    :param func_params: func的参数
    :param cwd_path: node_modules文件所在路径，如果不指定将使用全局的node_modules
    :param cmd_path: node的位置， 例如：r'D:\\path\to\node.exe'
    :return: func_name函数的执行结果
    """
    import execjs

    if func_params is None:
        func_params = []
    name = "MyNode"
    execjs.register(name, execjs._external_runtime.ExternalRuntime(
        name="Node.js (V8)",
        command=[cmd_path],
        encoding='UTF-8',
        runner_source=execjs._runner_sources.Node
    ))
    with open(js_file_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    js = ''.join(lines)
    js_context = execjs.get(name).compile(js, cwd=cwd_path)
    return js_context.call(func_name, *func_params)


def obsolete_setter(setter, attrname):
    def newsetter(self, value):
        c = self.__class__.__name__
        msg = "%s.%s is not modifiable, use %s.replace() instead" % (c, attrname, c)
        raise AttributeError(msg)
    return newsetter


def monkey_patch_json():
    json.__name__ = 'ujson'
    json.dumps = ujson.dumps
    json.loads = ujson.loads


def install_patch_method():
    monkey_patch_json()



from aioscpy.middleware.downloader import DownloaderMiddlewareManager
from aioscpy.middleware.itempipeline import ItemPipelineManager
from aioscpy.middleware.extension import ExtensionManager


__all__ = [
    "DownloaderMiddlewareManager",
    "ItemPipelineManager",
    "ExtensionManager",
]



from ._queue import spider_priority_queue, priority_queue
from ._queue_async import spider_aio_priority_queue, aio_priority_queue


__all__ = [
    spider_priority_queue,
    priority_queue,
    spider_aio_priority_queue,
    aio_priority_queue,
]



import asyncio
import random

from datetime import datetime
from collections import deque

from aioscpy import signals
from aioscpy import call_grace_instance


class Slot:
    """Downloader slot"""

    def __init__(self, concurrency, randomize_delay, delay=0):
        self.concurrency = concurrency
        self.delay = delay
        self.randomize_delay = randomize_delay

        self.active = set()
        self.queue = deque()
        self.transferring = set()
        self.lastseen = 0
        self.delay_run = False

    def free_transfer_slots(self):
        return self.concurrency - len(self.transferring)

    def download_delay(self):
        if self.randomize_delay:
            return random.uniform(0.5 * self.delay, 1.5 * self.delay)
        return self.delay

    def close(self):
        self.delay_run = True

    def __repr__(self):
        cls_name = self.__class__.__name__
        return "%s(concurrency=%r, delay=%0.2f, randomize_delay=%r)" % (
            cls_name, self.concurrency, self.delay, self.randomize_delay)

    def __str__(self):
        return (
                "<downloader.Slot concurrency=%r delay=%0.2f randomize_delay=%r "
                "len(active)=%d len(queue)=%d len(transferring)=%d lastseen=%s>" % (
                    self.concurrency, self.delay, self.randomize_delay,
                    len(self.active), len(self.queue), len(self.transferring),
                    datetime.fromtimestamp(self.lastseen).isoformat()
                )
        )


class Downloader(object):
    DOWNLOAD_SLOT = 'download_slot'

    def __init__(self, crawler):
        self.settings = crawler.settings
        self.crawler = crawler
        self.slot = None
        self.active = set()
        self.call_helper = self.di.get("tools").call_helper
        self.handlers = call_grace_instance('downloader_handler', self.settings, crawler)
        self.total_concurrency = self.settings.getint('CONCURRENT_REQUESTS')
        self.domain_concurrency = self.settings.getint('CONCURRENT_REQUESTS_PER_DOMAIN')
        self.ip_concurrency = self.settings.getint('CONCURRENT_REQUESTS_PER_IP')
        self.randomize_delay = self.settings.getbool('RANDOMIZE_DOWNLOAD_DELAY')
        self.delay = self.settings.getfloat('DOWNLOAD_DELAY')
        self.middleware = call_grace_instance(self.di.get('downloader_middleware'), only_instance=True).from_crawler(crawler)
        self.process_queue_task = None
        self.engine = None

        crawler.signals.connect(self.close, signals.engine_stopped)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    async def open(self, spider, engine):
        conc = self.ip_concurrency if self.ip_concurrency else self.domain_concurrency
        self.slot = Slot(conc, self.randomize_delay, self.delay)
        self.engine = engine
        self.process_queue_task = asyncio.create_task(self._process_queue(spider, self.slot))

    async def fetch(self, request):
        self.active.add(request)
        self.slot.active.add(request)
        self.slot.queue.append(request)

    async def _process_queue(self, spider, slot):
        while True:
            await asyncio.sleep(0.1)
            while slot.queue and slot.free_transfer_slots() > 0:
                request = slot.queue.popleft()
                asyncio.create_task(self._download(slot, request, spider))
                slot.transferring.add(request)
                slot.active.remove(request)
                self.active.remove(request)
                if slot.download_delay():
                    await asyncio.sleep(slot.download_delay())

    async def _download(self, slot, request, spider):
        try:
            response = None
            response = await self.middleware.process_request(spider, request)
            process_request_method = getattr(spider, "process_request", None)
            if process_request_method:
                response = await self.call_helper(process_request_method, request)
            if response is None or isinstance(response, self.di.get('request')):
                request = response or request
                response = await self.handlers.download_request(request, spider)
        except (Exception, BaseException, asyncio.TimeoutError) as exc:
            response = await self.middleware.process_exception(spider, request, exc)
            process_exception_method = getattr(spider, "process_exception", None)
            if process_exception_method:
                response = await self.call_helper(process_exception_method, request, exc)
        else:
            try:
                response = await self.middleware.process_response(spider, request, response)
                process_response_method = getattr(spider, "process_response", None)
                if process_response_method:
                    response = await self.call_helper(process_response_method, request, response)
            except (Exception, BaseException) as exc:
                response = exc
        finally:
            slot.transferring.discard(request)
            if isinstance(response, self.di.get('response')):
                response.request = request
            await self.engine._handle_downloader_output(response, request, spider)

    async def close(self):
        try:
            if self.slot is not None:
                self.slot.close()
            await self.handlers.close()
            if self.process_queue_task:
                self.process_queue_task.cancel()
        except (asyncio.CancelledError, Exception, BaseException) as exc:
            pass

    def needs_backout(self):
        return len(self.active) >= self.total_concurrency



from aioscpy.middleware.manager import MiddlewareManager
from aioscpy.utils.common import build_component_list


class ExtensionManager(MiddlewareManager):

    component_name = 'extension'

    @classmethod
    def _get_mwlist_from_settings(cls, settings):
        return build_component_list(settings.getwithbase('EXTENSIONS'))



from aioscpy.spider import Spider
from aioscpy import call_grace_instance


class HttpBinPostSpider(Spider):
    name = 'httpbin_post'
    custom_settings = {
        'CONCURRENT_REQUESTS': 10
    }

    start_urls = ['http://httpbin.org/post' for _ in range(20)]

    async def start_requests(self):
        """
        : request usage description:
        <form_request>: data = body
            [header]: Content-Type: application/x-www-form-urlencoded
            [method]: GET or POST
            [body]: a=1&b=1 or 
                  {
                    'a': 1,
                    'b': 2
                  }
        # supported special scenarios about json request
        <json_request>: json = body
            [header]: Content-Type: application/json
            [method]: POST
            [body]: {
                'a': 1,
                'b': 2
            }
        """
        for url in self.start_urls:
            yield call_grace_instance(
                # self.di.get('form_request'), 
                self.di.get('json_request'), 
                url, 
                method='POST',
                formdata={"b": '11'}
            )

    async def parse(self, response):
        item = await response.json
        yield item

    async def process_item(self, item):
        self.logger.info(item)


if __name__ == '__main__':
    q = HttpBinPostSpider()
    q.start()



import pprint
import asyncio
import anyio
import signal
import traceback

from typing import Optional, Type, Union, Any

from aioscpy.settings import overridden_settings
from aioscpy.settings import Settings
from aioscpy.signalmanager import SignalManager
from aioscpy.utils.ossignal import install_shutdown_handlers, signal_names
from aioscpy.inject import DependencyInjection
from aioscpy import call_grace_instance
from aioscpy.spider import Spider


class Crawler(object):

    def __init__(self, spidercls: Type[Spider], settings: Union[Settings, dict, None] = None, * args, ** kwargs):

        if isinstance(settings, dict) or settings is None:
            settings = Settings(settings)

        self.spidercls = spidercls
        self.settings = settings.copy()
        self.spidercls.update_settings(self.settings)
        self.signals = SignalManager(self)

        if d := dict(overridden_settings(self.settings)):
            self.logger.info("Overridden settings {spider}:\n{settings}",
                        **{'settings': pprint.pformat(d), "spider": spidercls.__name__})

        self.settings.freeze()
        self.crawling = False
        self.spider = self._create_spider(*args, **kwargs)
        self.engine = None
        self.stats = call_grace_instance('stats', self)
        self.DI = self._create_dependency()
        self.extensions = self.load('extension')
        self._close_wait = None

    async def crawl(self) -> None:
        if self.crawling:
            raise RuntimeError("Crawling already taking place")
        self.crawling = True

        try:
            if not self.spider.start_urls and "memory" in self.settings['SCHEDULER']:
                raise self.di.get("exceptions").UsageError("Crawler not found task exists in spider.start_urls.")
            await self.DI.inject_runner()
            self.engine = self._create_engine()
            start_requests = await self.di.get("tools").async_generator_wrapper(self.spider.start_requests())
            await self.engine.start(self.spider, start_requests)
            await self.di.get("tools").task_await(self, "_close_wait")
        except Exception as e:
            self.logger.error(f"crawler: {traceback.format_exc()}")
            self.crawling = False
            if self.engine is not None:
                await self.engine.close()
            raise e

    def load(self, key: str) -> Any:
        return self.DI.load(key)

    def _create_spider(self, *args, **kwargs):
        return self.spidercls.from_crawler(self, *args, **kwargs)

    def _create_engine(self):
        return call_grace_instance('engine', self, self.stop)

    def _create_dependency(self):
        return DependencyInjection(self.settings, self)

    async def stop(self):
        if self.crawling:
            self.crawling = False
            await self.engine.stop()
        self._close_wait = True


class CrawlerProcess(object):
    crawlers = property(
        lambda self: self._crawlers,
        doc="Set of :class:`crawlers <aioscpy.crawler.Crawler>`"
    )

    def __init__(self, settings: Union[Settings, dict, None] = None):
        if isinstance(settings, dict) or settings is None:
            settings = Settings(settings)
        self.settings = settings
        self._crawlers = set()
        self._active = set()
        self.bootstrap_failed = False
        self._group = []
        install_shutdown_handlers(self._signal_shutdown)
        self.di.get("log").std_log_aioscpy_info(settings)

    def crawl(self, crawler_or_spidercls: Union[Type[Spider], Crawler], *args, **kwargs) -> Crawler:
        crawler = self.create_crawler(crawler_or_spidercls, *args, **kwargs)
        self.crawlers.add(crawler)
        return crawler

    def crawl_soon(self, crawler_or_spidercls: Union[Type[Spider], Crawler], *args, **kwargs):
        crawler = self.crawl(crawler_or_spidercls, *args, **kwargs)
        self.active_crawler(crawler)

    def create_crawler(self, crawler_or_spidercls: Union[Type[Spider], Crawler], *args, **kwargs) -> Crawler:
        if isinstance(crawler_or_spidercls, Crawler):
            return crawler_or_spidercls
        if isinstance(crawler_or_spidercls, str):
            crawler_or_spidercls = self.load_spider(spider_key=crawler_or_spidercls, **kwargs)
        settings = kwargs.pop('settings', self.settings)
        return call_grace_instance("crawler", crawler_or_spidercls, *args, settings=settings, **kwargs)

    def active_crawler(self, crawler: Crawler):
        task = asyncio.create_task(crawler.crawl())
        self._active.add(task)

        def _done(result):
            self.crawlers.discard(crawler)
            self._active.discard(task)
            self.bootstrap_failed |= not getattr(crawler, 'spider', None)
            return result

        task.add_done_callback(_done)

    def load_spider(self, path: str = None, spider_key: str = None, spider_like: str = ""):
        if path is None:
            path = ''.join(['./', self.settings.get("NEWSPIDER_MODULE", "spiders")])
        spiders_cls = DependencyInjection.load_all_spider(path)

        if spider_key is not None:
            if spiders_cls.get(spider_key):
                return spiders_cls[spider_key]
            else:
                raise KeyError(f"Spider not found: {spider_key}")
        for name, spider_cls in spiders_cls.items():
            if spider_like and not spider_like in name:
                continue
            self.crawl(spider_cls)
            self.logger.debug(
                "Loading spider({name}) from {path}", **{"name": name, "path": path})

    async def stop(self):
        return await asyncio.gather(*[c.stop() for c in list(self.crawlers)])

    async def run(self):
        for crawler in self.crawlers:
            self.active_crawler(crawler)
        while self._active:
            self._group.append(await asyncio.gather(*self._active, return_exceptions=True))

    async def _graceful_stop_reactor(self):
        for task in self._active:
            task.cancel()
        for group in self._group:
            group.cancel()
        current_task = asyncio.all_tasks()
        for ct in current_task:
            ct.cancel()
        await asyncio.sleep(1)
        await self.stop()

    async def _force_stop_reactor(self):
        asyncio.get_running_loop().stop()

    def _signal_shutdown(self, signum, _):
        install_shutdown_handlers(self._signal_kill)
        signame = signal_names[signum]
        self.logger.info("Received {signame}, shutting down gracefully. Send again to force ",
                    **{'signame': signame})
        asyncio.create_task(self._graceful_stop_reactor())

    def _signal_kill(self, signum, _):
        install_shutdown_handlers(signal.SIG_IGN)
        signame = signal_names[signum]
        self.logger.info('Received {signame} twice, forcing unclean shutdown',
                    **{'signame': signame})
        asyncio.create_task(self._force_stop_reactor())

    def start(self):
        self.di.get("tools").install_event_loop_tips()
        self.di.get("tools").install_patch_method()
        try:
            anyio.run(self.run, backend='asyncio')
            # asyncio.run(self.run())
        except asyncio.CancelledError:
            pass



from urllib.parse import urlunparse

from aioscpy.exceptions import NotConfigured
from aioscpy.utils.tools import to_bytes
from aioscpy.utils.othtypes import urlparse_cached


def global_object_name(obj):
    return f"{obj.__module__}.{obj.__name__}"


def request_httprepr(request: "Request") -> bytes:
    """Return the raw HTTP representation (as bytes) of the given request.
    This is provided only for reference since it's not the actual stream of
    bytes that will be send when performing the request (that's controlled
    by Twisted).
    """
    parsed = urlparse_cached(request)
    path = urlunparse(('', '', parsed.path or '/', parsed.params, parsed.query, ''))
    s = to_bytes(request.method) + b" " + to_bytes(path) + b" HTTP/1.1\r\n"
    s += b"Host: " + to_bytes(parsed.hostname or b'') + b"\r\n"
    if request.headers:
        s += request.headers.to_string() + b"\r\n"
    s += b"\r\n"
    s += str(request.body).encode() if isinstance(request.body, dict) else request.body
    return s


def get_header_size(headers):
    size = 0
    for key, value in headers.items():
        if isinstance(value, (list, tuple)):
            for v in value:
                size += len(b": ") + len(key) + len(v)
    return size + len(b'\r\n') * (len(headers.keys()) - 1)


class DownloaderStats:

    def __init__(self, stats):
        self.stats = stats

    @classmethod
    def from_crawler(cls, crawler):
        if not crawler.settings.getbool('DOWNLOADER_STATS'):
            raise NotConfigured
        return cls(crawler.stats)

    def process_request(self, request, spider):
        self.stats.inc_value('downloader/request_count', spider=spider)
        self.stats.inc_value(f'downloader/request_method_count/{request.method}', spider=spider)
        reqlen = len(request_httprepr(request))
        self.stats.inc_value('downloader/request_bytes', reqlen, spider=spider)

    def process_response(self, request, response, spider):
        self.stats.inc_value('downloader/response_count', spider=spider)
        self.stats.inc_value(f'downloader/response_status_count/{response.status}', spider=spider)
        reslen = len(response.body) + get_header_size(response.headers) + 4
        # response.body + b"\r\n"+ response.header + b"\r\n" + response.status
        self.stats.inc_value('downloader/response_bytes', reslen, spider=spider)
        return response

    def process_exception(self, request, exception, spider):
        ex_class = global_object_name(exception.__class__)
        self.stats.inc_value('downloader/exception_count', spider=spider)
        self.stats.inc_value(f'downloader/exception_type_count/{ex_class}', spider=spider)



import asyncio

from aioscpy.spider import Spider


class HttpBinSpider(Spider):
    name = 'httpbin'
    custom_settings = {
        'CONCURRENT_REQUESTS': 10
    }
    start_urls = [f'http://httpbin.org/get?a{i}' for i in range(20)]

    async def parse(self, response):
        item = await response.json
        await asyncio.sleep(2)
        yield item

    async def process_item(self, item):
        pass
        # self.logger.info(item)


if __name__ == '__main__':
    q = HttpBinSpider()
    q.start()



from asyncio import Queue

from aioscpy.queue import BaseQueue


class PriorityQueue(BaseQueue):

    def __init__(self, server, spider, serializer="pickle"):
        super().__init__(server, spider)
        self.serializer = self.__compat__[serializer]

    def qsize(self) -> int:
        """Return the length of the queue"""
        return self.server.qsize()

    async def push(self, request):
        data = self._encode_request(request)
        await self.server.put(data)

    async def pop(self, timeout: int = 0, count: int = 0) -> list:
        _item = await self.server.get()
        return [self._decode_request(_item)]


def memory_queue(spider) -> PriorityQueue:
    """
    async def run():
        queue = memery_queue('message:queue')
        await queue.push({"url": "https://www.baidu.com/?kw=1", "task_id": '123'})
        print(await queue.pop())


    if __name__ == "__main__":
        import asyncio
        asyncio.run(run())

    """
    server = Queue()
    return PriorityQueue(server=server, spider=spider)


spider_queue = memory_queue



from __future__ import absolute_import, unicode_literals

import sys
import socket
import warnings
import aioscpy

from loguru import logger

from aioscpy.exceptions import AioscpyDeprecationWarning
# from aioscpy.settings import Settings
from aioscpy.utils.tools import get_project_settings


def set_log_config(formatter: str, settings):
    _log_config = {
        "default": {
            "handlers": [
                {
                    "sink": sys.stdout,
                    "format": formatter,
                    "level": settings.get('LOG_LEVEL', "TRACE")
                }
            ],
            "extra": {
                "host": socket.gethostbyname(socket.gethostname()),
                'log_name': settings.get("BOT_NAME", 'default'),
                'type': 'None'
            },
            "levels": [
                dict(name="TRACE", icon="✏️", color="<cyan><bold>"),
                dict(name="DEBUG", icon="❄️", color="<blue><bold>"),
                dict(name="INFO", icon="♻️", color="<bold>"),
                dict(name="SUCCESS", icon="✔️", color="<green><bold>"),
                dict(name="WARNING", icon="⚠️", color="<yellow><bold>"),
                dict(name="ERROR", icon="❌️", color="<red><bold>"),
                dict(name="CRITICAL", icon="☠️", color="<RED><bold>"),
            ]
        }
    }
    if settings.get('LOG_FILE', False):
        _log_config['default']['handlers'].append({
            "sink": settings.get('LOG_FILENAME', __file__),
            "format": formatter,
            "level": settings.get('LOG_LEVEL', "DEBUG"),
            "rotation": settings.get("LOG_ROTATION", '1 week'),
            "retention": settings.get("LOG_RETENTION", '30 days'),
            'encoding': settings.get("LOG_ENCODING", "utf-8")
        })
    return _log_config


class LogFormatter(object):
    simple_formatter = '<green>{time:YYYY-MM-DD HH:mm:ss}</green> ' \
                       '[<cyan>{name}</cyan>] ' \
                       '<level>{level.icon}{level}</level>: ' \
                       '<level>{message}</level> '

    default_formatter = '<green>{time:YYYY-MM-DD HH:mm:ss,SSS}</green> | ' \
                        '[<cyan>{extra[log_name]}</cyan>] <cyan>{module}</cyan>:<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | ' \
                        '<red>{extra[host]}</red> | ' \
                        '<level>{level.icon}{level: <5}</level> | ' \
                        '<level>{level.no}</level> | ' \
                        '<level>{extra[type]}</level> | ' \
                        '<level>{message}</level> '

    kafka_formatter = '{time:YYYY-MM-DD HH:mm:ss,SSS}| ' \
                      '[{extra[log_name]}] {module}:{name}:{function}:{line} | ' \
                      '{extra[host]} | ' \
                      '{process} | ' \
                      '{thread} | ' \
                      '{level: <5} | ' \
                      '{level.no} | ' \
                      '{extra[type]}| ' \
                      '{message} '

    @classmethod
    def setter_log_handler(cls, log, callback=None):
        assert callable(callback), 'callback must be a callable object'
        log.add(callback, format=cls.kafka_formatter)

    @classmethod
    def get_logger(cls, log, name=None):
        settings = get_project_settings()
        log_config = set_log_config(cls.simple_formatter, settings)
        config = log_config.pop('default', {})
        if name:
            config['extra']['log_name'] = name
        log.configure(**config)
        return log

    @staticmethod
    def format(spider, meta):
        if hasattr(spider, 'logging_keys'):
            logging_txt = []
            for key in spider.logging_keys:
                if meta.get(key, None) is not None:
                    logging_txt.append(u'{0}:{1} '.format(key, meta[key]))
            logging_txt.append('successfully')
            return ' '.join(logging_txt)


def logformatter_adapter(logkws):
    if not {'level', 'msg', 'args'} <= set(logkws):
        warnings.warn('Missing keys in LogFormatter method',
                      AioscpyDeprecationWarning)

    if 'format' in logkws:
        warnings.warn('`format` key in LogFormatter methods has been '
                      'deprecated, use `msg` instead',
                      AioscpyDeprecationWarning)

    level = logkws.get('level', 'INFO')
    message = logkws.get('format', logkws.get('msg'))
    args = logkws if not logkws.get('args') else logkws['args']

    return level, message, args


def std_log_aioscpy_info(settings):
    from pprint import pprint, pformat

    icon = """                          
         (_)                           
     __ _ _  ___  ___  ___ _ __  _   _ 
    / _` | |/ _ \/ __|/ __| '_ \| | | |
   | (_| | | (_) \__ \ (__| |_) | |_| |
    \__,_|_|\___/|___/\___| .__/ \__, |
                          | |     __/ |
                          |_|    |___/ 
    """
    logger.info("{item}", **{'item': icon})
    logger.info("aioscpy {version} started (bot: {bot})",
                **{'version': aioscpy.__version__, 'bot': settings['BOT_NAME']})


lof = LogFormatter

logger = lof.get_logger(logger)



import pprint

from asyncio import iscoroutinefunction
from collections import defaultdict, deque

from aioscpy.exceptions import NotConfigured


class MiddlewareManager:
    """Base class for implementing middleware managers"""

    component_name = 'foo middleware'

    def __init__(self, crawler=None, middlewares=None):
        self.crawler = crawler
        self.middlewares = middlewares
        self.methods = defaultdict(deque)
        for mw in middlewares:
            self._add_middleware(mw)

    @classmethod
    def _get_mwlist_from_settings(cls, settings):
        raise NotImplementedError

    @classmethod
    def from_settings(cls, settings, crawler=None):
        mwlist = cls._get_mwlist_from_settings(settings)
        middlewares = []
        enabled = []
        for clspath in mwlist:
            try:
                mw = crawler.DI.load_object_slot(clspath.split('.')[-2], clspath)
                middlewares.append(mw)
                enabled.append(clspath)
            except NotConfigured as e:
                if e.args:
                    clsname = clspath.split('.')[-1]
                    cls.logger.warning("Disabled {clsname}: {eargs}",
                                   **{'clsname': clsname, 'eargs': e.args[0]},
                                   extra={'crawler': crawler})
        if enabled:
            cls.logger.info("Enabled {name} {componentname}s:\n{enabledlist}",
                        **{'componentname': cls.component_name,
                         'enabledlist': pprint.pformat(enabled),
                         'name': crawler.spider.name},
                        extra={'crawler': crawler})
        return cls(crawler=crawler, middlewares=middlewares)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings, crawler)

    def _add_middleware(self, mw):
        if hasattr(mw, 'open_spider'):
            self.methods['open_spider'].append(mw.open_spider)
        if hasattr(mw, 'close_spider'):
            self.methods['close_spider'].appendleft(mw.close_spider)

    async def _process_parallel(self, methodname, obj, *args):
        return await self.process_parallel(self.methods[methodname], obj, *args)

    async def _process_chain(self, methodname, obj, *args):
        return await self.process_chain(self.methods[methodname], obj, *args)

    async def _process_chain_both(self, cb_methodname, eb_methodname, obj, *args):
        return await self.process_chain_both(self.methods[cb_methodname],
                                             self.methods[eb_methodname], obj, *args)

    async def open_spider(self, spider):
        return await self._process_parallel('open_spider', spider)

    async def close_spider(self, spider):
        return await self._process_parallel('close_spider', spider)

    @staticmethod
    async def process_parallel(callbacks, input_, *a, **kw):
        for callback in callbacks:
            if iscoroutinefunction(callback):
                await callback(input_, *a, **kw)
            else:
                callback(input_, *a, **kw)

    @staticmethod
    async def process_chain(callbacks, input_, *a, **kw):
        for callback in callbacks:
            if iscoroutinefunction(callback):
                input_result = await callback(input_, *a, **kw)
            else:
                input_result = callback(input_, *a, **kw)
            if input_result is not None:
                input_ = input_result
        return input_

    @staticmethod
    async def process_chain_both(callbacks, errbacks, input_, *a, **kw):
        for cb, eb in zip(callbacks, errbacks):
            try:
                if iscoroutinefunction(cb):
                    input_ = await cb(input_, *a, **kw)
                else:
                    input_ = cb(input_, *a, **kw)
            except(Exception, BaseException) as e:
                if iscoroutinefunction(cb):
                    input_ = await eb(input_, *a, **kw)
                else:
                    input_ = eb(input_, *a, **kw)
            return input_



import asyncio
import ssl
import aiohttp

from anti_header import Headers
from anti_useragent.utils.cipers import generate_cipher


class AioHttpDownloadHandler(object):

    def __init__(self, settings, crawler):
        self.settings = settings
        self.crawler = crawler
        self.aiohttp_client_session = {
            'timeout': aiohttp.ClientTimeout(total=20),
            'trust_env': True,
            "connector": aiohttp.TCPConnector(
                verify_ssl=False,
                limit=1000,
                force_close=True,
                use_dns_cache=False,
                limit_per_host=200,
                enable_cleanup_closed=True
            )
        }
        self.session_stats = self.settings.getbool("REQUESTS_SESSION_STATS", False)
        self.session = None
        self.context = ssl.create_default_context()

    @classmethod
    def from_settings(cls, settings, crawler):
        return cls(settings, crawler)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings, crawler)

    async def download_request(self, request, spider):
        session_kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
        }
        if isinstance(request.body, dict):
            session_kwargs['json'] = request.body or None
        else:
            session_kwargs['data'] = request.body or None
        
        headers = request.headers
        if isinstance(headers, Headers):
            headers = headers.to_unicode_dict()
        session_kwargs['headers'] = headers

        if request.meta.get('TLS_CIPHERS') or self.settings.get('TLS_CIPHERS'):
            self.context.set_ciphers(generate_cipher())
            session_kwargs['ssl'] = self.context

        if request.meta.get("proxy"):
            session_kwargs["proxy"] = request.meta['proxy']
            self.logger.debug(f"use {request.meta['proxy']} crawling: {request.url}")

        if self.session_stats:
            if self.session is None:
                self.session = aiohttp.ClientSession(**self.aiohttp_client_session)
            response = await self.session.request(request.method, request.url, **session_kwargs)
            content = await response.read()
        else:
            async with aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=20),
                trust_env=True,
                connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
                async with session.request(request.method, request.url, **session_kwargs) as response:
                    content = await response.read()

        return self.di.get("response")(
            str(response.url),
            status=response.status,
            headers=response.headers,
            body=content,
            cookies=response.cookies,
            _response=response)

    async def close(self):
        if self.session is not None:
            await self.session.close()

        # Wait 250 ms for the underlying SSL connections to close
        # https://docs.aiohttp.org/en/latest/client_advanced.html#graceful-shutdown
        await asyncio.sleep(0.250)



from aioscpy import signals
from aioscpy import call_grace_instance


class Spider(object):
    name = None
    custom_settings = None

    def __init__(self, name=None, **kwargs):
        if name is not None:
            self.name = name
        self.__dict__.update(kwargs)
        if not hasattr(self, 'start_urls'):
            self.start_urls = []

    def log(self, message, level='DEBUG', **kw):
        self.logger.log(level, message, **kw)

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        return spider

    def _set_crawler(self, crawler):
        self.crawler = crawler
        self.settings = crawler.settings
        crawler.signals.connect(self.close, signals.spider_closed)
        crawler.signals.connect(self.spider_idle, signal=signals.spider_idle)

    async def start_requests(self):
        for url in self.start_urls:
            yield self.di.get('request')(url, dont_filter=True)

    async def _parse(self, response, **kwargs):
        return self.parse(response)

    async def parse(self, response):
        raise NotImplementedError(f'{self.__class__.__name__}.parse callback is not defined')

    @classmethod
    def update_settings(cls, settings):
        settings.setdict(cls.custom_settings or {}, priority='spider')

    @staticmethod
    def close(spider, reason):
        closed = getattr(spider, 'closed', None)
        if callable(closed):
            return closed(reason)

    @classmethod
    def start(cls):
        from aioscpy.crawler import CrawlerProcess
        from aioscpy.utils.tools import get_project_settings

        process = call_grace_instance(CrawlerProcess, get_project_settings())
        process.crawl(cls)
        process.start()

    def spider_idle(self):
        if self.settings.get("SPIDER_IDLE", True):
            raise self.di.get('exceptions').DontCloseSpider

    def __str__(self):
        return "<%s %r at 0x%0x>" % (type(self).__name__, self.name, id(self))

    __repr__ = __str__


Spider = call_grace_instance('spider', only_instance=True)



from contextlib import suppress

import re
import parsel
from w3lib.encoding import (html_body_declared_encoding, html_to_unicode,
                            http_content_type_encoding, resolve_encoding)
from w3lib.html import strip_html5_whitespace
from parsel import Selector

from aioscpy.http import Request
from aioscpy.http.response import Response
from aioscpy.utils.tools import to_unicode, memoizemethod_noargs, call_helper


class TextResponse(Response):

    _DEFAULT_ENCODING = 'ascii'

    def __init__(self, *args, **kwargs):
        self._encoding = kwargs.pop('encoding', None) or "utf-8"
        self._cached_benc = None
        self._cached_ubody = None
        self._cached_selector = None
        self.cookies = self._set_cookies(kwargs.pop("cookies", None))
        super(TextResponse, self).__init__(*args, **kwargs)

    @staticmethod
    def _set_cookies(cookies_raw):
        cookies = {}
        if cookies_raw is None:
            return cookies
        cookies_str = str(cookies_raw)
        for cookie in re.findall(r'Set-Cookie: (.*?)=(.*?); Domain', cookies_str, re.S):
            cookies[cookie[0]] = cookie[1]
        return cookies

    def _set_url(self, url):
        if isinstance(url, str):
            self._url = to_unicode(url, self.encoding)
        else:
            super(TextResponse, self)._set_url(url)

    def _set_body(self, body):
        self._body = b''  # used by encoding detection
        if isinstance(body, str):
            if self._encoding is None:
                raise TypeError('Cannot convert unicode body - %s has no encoding' %
                                type(self).__name__)
            self._body = body.encode(self._encoding)
        else:
            super(TextResponse, self)._set_body(body)

    def replace(self, *args, **kwargs):
        kwargs.setdefault('encoding', self.encoding)
        return Response.replace(self, *args, **kwargs)

    @property
    def encoding(self):
        return self._declared_encoding() or self._body_inferred_encoding()

    def _declared_encoding(self):
        return self._encoding or self._headers_encoding() \
            or self._body_declared_encoding()

    def body_as_unicode(self):
        """Return body as unicode"""
        return self.text

    @property
    def text(self):
        """ Body as unicode """
        # access self.encoding before _cached_ubody to make sure
        # _body_inferred_encoding is called
        benc = self.encoding
        if self._cached_ubody is None:
            charset = 'charset=%s' % benc
            self._cached_ubody = html_to_unicode(charset, self.body)[1]
        return self._cached_ubody

    @property
    async def json(self):
        return await call_helper(self._response.json)

    @memoizemethod_noargs
    def _headers_encoding(self):
        content_type = self.headers.get(b'Content-Type', b'')
        return http_content_type_encoding(to_unicode(content_type))

    def _body_inferred_encoding(self):
        if self._cached_benc is None:
            content_type = to_unicode(self.headers.get(b'Content-Type', b''))
            benc, ubody = html_to_unicode(content_type, self.body,
                                          auto_detect_fun=self._auto_detect_fun,
                                          default_encoding=self._DEFAULT_ENCODING)
            self._cached_benc = benc
            self._cached_ubody = ubody
        return self._cached_benc

    def _auto_detect_fun(self, text):
        for enc in (self._DEFAULT_ENCODING, 'utf-8', 'cp1252'):
            try:
                text.decode(enc)
            except UnicodeError:
                continue
            return resolve_encoding(enc)

    @memoizemethod_noargs
    def _body_declared_encoding(self):
        return html_body_declared_encoding(self.body)

    @property
    def selector(self):
        if self._cached_selector is None:
            text = self.text
            self._cached_selector = Selector(text)
        return self._cached_selector

    def xpath(self, query, **kwargs):
        return self.selector.xpath(query, **kwargs)

    def css(self, query):
        return self.selector.css(query)

    def follow(self, url, callback=None, method='GET', headers=None, body=None,
               cookies=None, meta=None, encoding=None, priority=0,
               dont_filter=False, errback=None, cb_kwargs=None, flags=None, **kwargs):
        if isinstance(url, parsel.Selector):
            url = _url_from_selector(url)
        elif isinstance(url, parsel.SelectorList):
            raise ValueError("SelectorList is not supported")
        encoding = self.encoding if encoding is None else encoding
        return super(TextResponse, self).follow(
            url=url,
            callback=callback,
            method=method,
            headers=headers,
            body=body,
            cookies=cookies,
            meta=meta,
            encoding=encoding,
            priority=priority,
            dont_filter=dont_filter,
            errback=errback,
            cb_kwargs=cb_kwargs,
            flags=flags,
            **kwargs
        )

    def follow_all(self, urls=None, callback=None, method='GET', headers=None, body=None,
                   cookies=None, meta=None, encoding=None, priority=0,
                   dont_filter=False, errback=None, cb_kwargs=None, flags=None,
                   css=None, xpath=None):
        arguments = [x for x in (urls, css, xpath) if x is not None]
        if len(arguments) != 1:
            raise ValueError(
                "Please supply exactly one of the following arguments: urls, css, xpath"
            )
        if not urls:
            if css:
                urls = self.css(css)
            if xpath:
                urls = self.xpath(xpath)
        if isinstance(urls, parsel.SelectorList):
            selectors = urls
            urls = []
            for sel in selectors:
                with suppress(_InvalidSelector):
                    urls.append(_url_from_selector(sel))
        return super(TextResponse, self).follow_all(
            urls=urls,
            callback=callback,
            method=method,
            headers=headers,
            body=body,
            cookies=cookies,
            meta=meta,
            encoding=encoding,
            priority=priority,
            dont_filter=dont_filter,
            errback=errback,
            cb_kwargs=cb_kwargs,
            flags=flags,
        )


class _InvalidSelector(ValueError):
    """
    Raised when a URL cannot be obtained from a Selector
    """


def _url_from_selector(sel):
    # type: (parsel.Selector) -> str
    if isinstance(sel.root, str):
        # e.g. ::attr(href) result
        return strip_html5_whitespace(sel.root)
    if not hasattr(sel.root, 'tag'):
        raise _InvalidSelector("Unsupported selector: %s" % sel)
    if sel.root.tag not in ('a', 'link'):
        raise _InvalidSelector("Only <a> and <link> elements are supported; got <%s>" %
                               sel.root.tag)
    href = sel.root.get('href')
    if href is None:
        raise _InvalidSelector("<%s> element has no href attribute: %s" %
                               (sel.root.tag, sel))
    return strip_html5_whitespace(href)



import pika

from aioscpy.queue import BaseQueue


class PriorityQueue(BaseQueue):
    def qsize(self) -> int:
        return self.server.get_waiting_message_count()

    def push(self, request: dict):
        data = self._encode_request(request)
        score = request.get('priority', 1)

        self.server.basic_publish(
            properties=pika.BasicProperties(priority=score),
            exchange='',
            routing_key=self.key,
            body=data
        )

    def on_message(self, ch, method, properties, body):
        pass

    def m_pop(self, on_message_callback=None, auto_ack=False):
        if not on_message_callback:
            on_message_callback = self.on_message
        self.server.basic_consume(
            on_message_callback=on_message_callback,
            queue=self.key,
            auto_ack=auto_ack
        )
        self.server.start_consuming()

    def pop(self, auto_ack=False):
        _method, _, _body = self.server.basic_get(queue=self.key, auto_ack=auto_ack)
        if all([isinstance(_body, bytes), _body is not None]):
            return _method, self._decode_request(_body)
        return None, None

    def finish(self, method):
        self.server.basic_ack(delivery_tag=method.delivery_tag)


class RabbitMq:
    __mq_instance = None
    __mq_connection_instance = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = self.validator(kwargs)

    @staticmethod
    def validator(params: dict) -> dict:
        params.setdefault('host', '127.0.0.1')
        params.setdefault('port', 5672)
        params.setdefault('username', 'admin')
        params.setdefault('password', 'admin')
        params.setdefault('max_priority', 100)
        params.setdefault('key', 'rabbitmq:queue')
        return params

    @property
    def get_channel(self):
        if not self.__mq_instance:
            connection = pika.BlockingConnection(
                pika.ConnectionParameters(
                    host=self.kwargs['host'],
                    port=self.kwargs['port'],
                    credentials=pika.PlainCredentials(
                        username=self.kwargs['username'],
                        password=self.kwargs['password']
                    )
                )
            )
            channel = connection.channel()
            channel.queue_declare(
                queue=self.kwargs['key'],
                arguments={"x-max-priority": self.kwargs['max_priority']}
            )
            self.__mq_instance, self.__mq_connection_instance = channel, connection
        return self.__mq_instance

    def close(self):
        if self.__mq_instance:
            self.__mq_instance.close()
            self.__mq_connection_instance.close()


def priority_queue(key: str, mq: dict) -> PriorityQueue:
    """
    # unit test example
    def run():
        queue = rabbitmq_client('message:queue')
        for i in range(5):
            queue.push({"url": f"https://www.baidu.com/?kw={i}", "task_id": '123'})
        while 1:
            method, msg = queue.pop()
            print(msg)
            if not msg:
                break
            time.sleep(1)
            if method:
                queue.finish(method)

    run()

    """
    server = RabbitMq(**mq).get_channel
    return PriorityQueue(server=server, key=key)


spider_priority_queue = priority_queue



from aioscpy.http.request import Request
from aioscpy.http.request.form import FormRequest
from aioscpy.http.request.json import JsonRequest

from aioscpy.http.response import Response
from aioscpy.http.response.text import TextResponse


__all__ = [
    Request,
    FormRequest,
    JsonRequest,
    Response,
    TextResponse
]



import asyncio


class Scheduler(object):

    def __init__(self, _queue_df, spider, stats):
        self.queue = _queue_df
        self.stats = stats
        self.spider = spider

    @classmethod
    def from_crawler(cls, crawler):
        raise NotImplementedError(
            '{} from_crawler method must define'.format(cls.__class__.__name__))

    async def enqueue_request(self, request):
        if self.stats:
            self.stats.inc_value('scheduler/enqueued/redis', spider=self.spider)
        await self.queue.push(request)
        return True

    async def async_next_request(self):
        _results = await self.queue.pop(count=100)
        self.stats.inc_value('scheduler/dequeued/redis', count=len(_results), spider=self.spider)
        return _results

    async def open(self, start_requests):
        if asyncio.iscoroutine(self.queue):
            self.queue = await self.queue
        async for request in start_requests:
            await self.enqueue_request(request)

    async def close(self, slot):
        if slot.inprogress:
            for request in slot.inprogress:
                await self.enqueue_request(request)
        await self.queue.close()

    def __len__(self):
        return self.queue.qsize()

    async def has_pending_requests(self):
        return len(self) > 0



from aioscpy.core.scheduler import Scheduler
from aioscpy.queue.memory import memory_queue


class MemoryScheduler(Scheduler):

    @classmethod
    def from_crawler(cls, crawler):
        return cls(_queue_df=memory_queue(crawler.spider), stats=crawler.stats, spider=crawler.spider)



import asyncio
import traceback

from collections import deque

from aioscpy import signals, call_grace_instance


class Slot:
    MIN_RESPONSE_SIZE = 1024

    def __init__(self, max_active_size: int = 5000):
        self.max_active_size = max_active_size
        self.queue = deque()
        self.active = set()
        self.active_size = 0
        self.itemproc_size = 0
        self.closing_future = None
        self.closing_lock = True

    def add_response_request(self, response, request):
        self.queue.append((response, request))
        self.active.add(request)
        if hasattr(response, 'body'):
            self.active_size += max(len(response.body), self.MIN_RESPONSE_SIZE)

    def next_response_request_deferred(self):
        response, request = self.queue.popleft()
        return response, request

    def finish_response(self, request, response):
        self.active.discard(request)
        # self.logger.warning(f'start finish response active del: {self.active_size}, active: {len(self.active)}, response: {len(response.body)}')
        if hasattr(response, 'body'):
            self.active_size -= max(len(response.body), self.MIN_RESPONSE_SIZE)
        request, response = None, None        

    def is_idle(self):
        return self.queue or self.active

    def needs_backout(self):
        return self.active_size > self.max_active_size


class Scraper:

    def __init__(self, crawler):
        self.slot = None
        self.itemproc = call_grace_instance(self.di.get('item_processor'), only_instance=True).from_crawler(crawler)
        self.crawler = crawler
        self.signals = crawler.signals
        self.logformatter = crawler.load("log_formatter")
        self.call_helper = self.di.get("tools").call_helper
        self.concurrent_items_semaphore = asyncio.Semaphore(crawler.settings.getint('CONCURRENT_ITEMS', 100))
        self.task_scrape_next = None

    async def open_spider(self, spider):
        self.slot = call_grace_instance(Slot, self.crawler.settings.getint('SCRAPER_SLOT_MAX_ACTIVE_SIZE', 500000))
        await self.itemproc.open_spider(spider)
        self.task_scrape_next = asyncio.create_task(self._scrape_next(spider, self.slot))

    async def close_spider(self, spider):
        slot = self.slot
        await self.itemproc.close_spider(spider)
        if self.task_scrape_next is not None:
            self.task_scrape_next.cancel()
        self._check_if_closing(spider, slot)

    def is_idle(self):
        return not self.slot

    def _check_if_closing(self, spider, slot):
        if not slot.closing_future and slot.is_idle() and slot.closing_lock:
            slot.closing_lock = False

    async def enqueue_scrape(self, response, request):
        slot = self.slot
        slot.add_response_request(response, request)

    async def _scrape_next(self, spider, slot):
        local_lock = True
        while True:
            while local_lock and slot.queue:
                local_lock = False
                response, request = slot.next_response_request_deferred()
                asyncio.create_task(self._scrape(response, request, spider))
                local_lock = True
            await asyncio.sleep(2)

    async def _scrape(self, result, request, spider):
        if not isinstance(result, (self.di.get('response'), Exception, BaseException)):
            raise TypeError(f"Incorrect type: expected Response or Failure, got {type(result)}: {result!r}")
        try:
            response = await self._scrape2(result, request, spider)  # returns spider's processed output
            await self.handle_spider_output(response, request, result, spider)
        except (Exception, BaseException) as e:
            await self.handle_spider_error(e, request, result, spider)
        self.slot.finish_response(request, result)
        return request, result

    async def _scrape2(self, result, request, spider):
        try:
            return await self.call_spider(result, request, spider)
        except (Exception, BaseException) as e:
            await self._log_download_errors(e, result, request, spider)

    async def call_spider(self, result, request, spider):
        if isinstance(result, self.di.get('response')):
            callback = request.callback or spider._parse
            result.request = request
            return await self.call_helper(callback, result, **result.request.cb_kwargs)
        else:
            if request.errback is None:
                raise result
            return await self.call_helper(request.errback, result)

    async def handle_spider_error(self, exc, request, response, spider):
        if isinstance(exc, self.di.get('exceptions').CloseSpider):
            asyncio.create_task(self.crawler.engine.close_spider(spider, exc.reason or 'cancelled'))
            return
        logkws = self.logformatter.spider_error(exc, request, response, spider)
        level, message, kwargs = self.di.get("log").logformatter_adapter(logkws)
        if type(exc).__name__ not in ['CancelledError']:
            self.logger.log(level, message, exception=True, depth=2, **kwargs)
            self.logger.error(f"handle_spider_error: {traceback.format_exc()}")
        await self.signals.send_catch_log(
            signal=signals.spider_error,
            failure=exc, response=response,
            spider=spider
        )
        self.crawler.stats.inc_value(
            "spider_exceptions/%s" % exc.__class__.__name__,
            spider=spider
        )

    async def handle_spider_output(self, result, request, response, spider):
        if not result:
            return

        while True:
            try:
                res = await result.__anext__()
            except StopAsyncIteration:
                break
            except Exception as e:
                await self.handle_spider_error(e, request, response, spider)
            else:
                await self._process_spidermw_output(res, request, response, spider)

    async def _process_spidermw_output(self, output, request, response, spider):
        async with self.concurrent_items_semaphore:
            if isinstance(output, self.di.get('request')):
                await self.crawler.engine.crawl(request=output, spider=spider)
            elif isinstance(output, dict):
                try:
                    self.slot.itemproc_size += 1
                    item = await self.itemproc.process_item(output, spider)
                    process_item_method = getattr(spider, 'process_item', None)
                    if process_item_method:
                        item = await self.call_helper(process_item_method, item)
                except Exception as e:
                    self.logger.error(f"process_spidermw_output: {traceback.format_exc()}")
                await self._itemproc_finished(output, item, request, response, spider)
            elif output is None:
                pass
            else:
                typename = type(output).__name__
                self.logger.error(
                    'Spider must return request, item, or None, got %(typename)r in %(request)s',
                    {'request': request, 'typename': typename},
                    extra={'spider': spider},
                )

    async def _log_download_errors(self, spider_exception, download_exception, request, spider):
        if isinstance(download_exception, (Exception, BaseException)) \
                and not isinstance(download_exception, self.di.get('exceptions').IgnoreRequest):
            logkws = self.logformatter.download_error(download_exception, request, spider)
            level, message, kwargs = self.di.get("log").logformatter_adapter(logkws)
            if type(download_exception).__name__ not in ['CancelledError']:
                self.logger.log(level, message, **kwargs)
                self.logger.error(traceback.format_exc())

        if spider_exception is not download_exception:
            raise spider_exception

    async def _itemproc_finished(self, output, item, request, response, spider):
        self.slot.itemproc_size -= 1
        if isinstance(output, (Exception, BaseException)):
            if isinstance(output, self.di.get('exceptions').DropItem):
                logkws = self.logformatter.dropped(item, output, response, spider)
                if logkws is not None:
                    level, message, kwargs = self.di.get("log").logformatter_adapter(logkws)
                    if type(output).__name__ not in ['CancelledError']:
                        self.logger.log(level, message, **kwargs)
                        self.logger.error(traceback.format_exc())
                await self.signals.send_catch_log_coroutine(
                    signal=signals.item_dropped, item=item, response=response,
                    spider=spider, exception=output)
            else:
                logkws = self.logformatter.item_error(item, output, response, spider)
                level, message, kwargs = self.di.get("log").logformatter_adapter(logkws)
                if type(output).__name__ not in ['CancelledError']:
                    self.logger.log(level, message, **kwargs)
                    self.logger.error(traceback.format_exc())
                await self.signals.send_catch_log_coroutine(
                    signal=signals.item_error, item=item, response=response,
                    spider=spider, failure=output)
        else:
            logkws = self.logformatter.scraped(output, response, spider)
            if logkws is not None:
                level, message, kwargs = self.di.get("log").logformatter_adapter(logkws)
                self.logger.log(level, message, **kwargs)
            await self.signals.send_catch_log_coroutine(
                signal=signals.item_scraped, item=output, response=response,
                spider=spider)



import numbers
import warnings
import os

from operator import itemgetter
from importlib import import_module

from aioscpy.settings import BaseSettings


def without_none_values(iterable):
    """Return a copy of ``iterable`` with all ``None`` entries removed.

    If ``iterable`` is a mapping, return a dictionary where all pairs that have
    value ``None`` have been removed.
    """
    try:
        return {k: v for k, v in iterable.items() if v is not None}
    except AttributeError:
        return type(iterable)((v for v in iterable if v is not None))


def build_component_list(compdict, custom=None):
    """Compose a component list from a { class: order } dictionary."""

    def _check_components(complist):
        if len({c for c in complist}) != len(complist):
            raise ValueError(f'Some paths in {complist!r} convert to the same object, '
                             'please update your settings')

    def _map_keys(compdict):
        if isinstance(compdict, BaseSettings):
            compbs = BaseSettings()
            for k, v in compdict.items():
                prio = compdict.getpriority(k)
                if compbs.getpriority(k) == prio:
                    raise ValueError(f'Some paths in {list(compdict.keys())!r} '
                                     'convert to the same '
                                     'object, please update your settings'
                                     )
                else:
                    compbs.set(k, v, priority=prio)
            return compbs
        else:
            _check_components(compdict)
            return {k: v for k, v in compdict.items()}

    def _validate_values(compdict):
        """Fail if a value in the components dict is not a real number or None."""
        for name, value in compdict.items():
            if value is not None and not isinstance(value, numbers.Real):
                raise ValueError(f'Invalid value {value} for component {name}, '
                                 'please provide a real number or None instead')

    # BEGIN Backward compatibility for old (base, custom) call signature
    if isinstance(custom, (list, tuple)):
        _check_components(custom)
        return type(custom)(c for c in custom)

    if custom is not None:
        compdict.update(custom)
    # END Backward compatibility

    _validate_values(compdict)
    compdict = without_none_values(_map_keys(compdict))
    return [k for k, v in sorted(compdict.items(), key=itemgetter(1))]


def arglist_to_dict(arglist):
    """Convert a list of arguments like ['arg1=val1', 'arg2=val2', ...] to a
    dict
    """
    return dict(x.split('=', 1) for x in arglist)


def inside_project():
    aioscpy_module = os.environ.get('AIOSCPY_SETTINGS_MODULE')
    if aioscpy_module is not None:
        try:
            import_module(aioscpy_module)
        except ImportError as exc:
            warnings.warn(f"Cannot import aioscpy settings module {aioscpy_module}: {exc}")
        else:
            return True
    return bool(closest_aioscpy_cfg())


def closest_aioscpy_cfg(path='.', prevpath=None):
    """Return the path to the closest aioscpy.cfg file by traversing the current
    directory and its parents
    """
    if path == prevpath:
        return ''
    path = os.path.abspath(path)
    cfgfile = os.path.join(path, 'aioscpy.cfg')
    if os.path.exists(cfgfile):
        return cfgfile
    return closest_aioscpy_cfg(os.path.dirname(path), path)



import pkgutil

from aioscpy.inject import call_grace_instance

__version__ = (pkgutil.get_data(__package__, "VERSION") or b"").decode("ascii").strip()

__all__ = [
    '__version__',
    'call_grace_instance'
]



import os
import sys
from shutil import rmtree
from os.path import dirname, join

from setuptools import setup, Command, find_packages

# Package meta-data.
NAME = "aioscpy"
DESCRIPTION = "An asyncio + aiolibs crawler imitate scrapy framework"
URL = "https://github.com/ihandmine/aioscpy"
EMAIL = "handmine@outlook.com"
AUTHOR = "handmine"
REQUIRES_PYTHON = ">=3.8.0"

here = os.path.abspath(os.path.dirname(__file__))
with open(f"{here}/README.md", encoding='utf-8') as f:
    long_description = f.read()

with open(join(dirname(__file__), 'aioscpy/VERSION'), 'rb') as f:
    old_version = f.read().decode('ascii').strip()
    maxv, midv, minv = [int(v) for v in old_version.split('.')]
    if minv <= 24:
        minv += 1
    else:
        midv += 1
        minv = 0
    VERSION = '.'.join([str(v) for v in [maxv, midv, minv]])
    print(f'old version: {old_version}, new version: {VERSION}')


class UploadCommand(Command):
    """Support setup_bak.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel distribution...")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        with open(join(dirname(__file__), 'aioscpy/VERSION'), 'w') as f:
            f.write(VERSION + '\n')

        self.status("git option [add]")
        os.system("git add aioscpy/VERSION")

        self.status("git option [commit][push]")
        os.system(f'git commit -m "{VERSION}"')
        os.system("git push")
        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    packages=find_packages(),
    include_package_data=True,
    package_data={"": ["*.py", "*.tmpl", '*.cfg']},
    install_requires=[
        "aiohttp",
        "httpx",
        "anti-header",
        "w3lib",
        "parsel",
        "PyDispatcher",
        "redis",
        "anyio"
    ],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=URL,
    author_email=EMAIL,
    license="MIT",
    keywords="""
                crawler
                scrapy
                asyncio
                aiohttp
                anti-header
                anti-useragent
                python3
               """,
    python_requires=REQUIRES_PYTHON,
    zip_safe=False,
    entry_points={
        'console_scripts': ['aioscpy = aioscpy.cmdline:execute']
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
    ],
    # Build and upload package: python3 setup_bak.py upload
    cmdclass={"upload": UploadCommand},
)



from aioscpy.queue.compat import COMPAT_TYPE

from aioscpy.queue.convert import request_from_dict, request_to_dict


class BaseQueue(object):

    __slots__ = ["server", "key", "serializer", "spider"]
    __compat__ = COMPAT_TYPE

    def __init__(self, server, spider=None, key=None, serializer=None):
        if serializer is None:
            serializer = self.__compat__[serializer or "json"]

        if not hasattr(serializer, 'loads'):
            raise TypeError("serializer does not implement 'loads' function: %r"
                            % serializer)
        if not hasattr(serializer, 'dumps'):
            raise TypeError("serializer does not implement 'dumps' function: %r"
                            % serializer)

        self.server = server
        self.key = key or 'sp:requests'
        self.serializer = serializer
        self.spider = spider

    def _encode_request(self, request) -> bytes:
        obj = request_to_dict(request, self.spider)
        return self.serializer.dumps(obj)

    def _decode_request(self, encoded_request: bytes) -> dict:
        obj = self.serializer.loads(encoded_request)
        return request_from_dict(obj, self.spider)
        # return obj

    def __len__(self):
        raise Exception('please use function len()')

    async def qsize(self):
        raise NotImplementedError

    async def push(self, request):
        raise NotImplementedError

    async def pop(self, timeout=0):
        raise NotImplementedError

    async def clear(self):
        await self.server.delete(self.key)

    async def close(self):
        if hasattr(self.server, "close"):
            await self.server.close()



import os

from aioscpy.utils.tools import referer_str


SCRAPEDMSG = "Scraped from {src}" + os.linesep + "{item}"
DROPPEDMSG = "Dropped: {exception}" + os.linesep + "{item}"
CRAWLEDMSG = "Crawled ({status}) {request}{request_flags} (referer: {referer}){response_flags}"
ITEMERRORMSG = "Error processing {item}"
SPIDERERRORMSG = "Spider error processing {request} (referer: {referer})"
DOWNLOADERRORMSG_SHORT = "Error downloading {request}"
DOWNLOADERRORMSG_LONG = "Error downloading {request}: {errmsg}"


class LogFormatter:

    def crawled(self, request, response, spider):
        request_flags = f' {str(request.flags)}' if request.flags else ''
        response_flags = f' {str(response.flags)}' if response.flags else ''
        return {
            'level': "DEBUG",
            'msg': CRAWLEDMSG,
            'args': {
                'status': response.status,
                'request': request,
                'request_flags': request_flags,
                'referer': referer_str(request),
                'response_flags': response_flags,
                # backward compatibility with Aioscpy logformatter below 1.4 version
                'flags': response_flags
            }
        }

    def scraped(self, item, response, spider):
        """Logs a message when an item is scraped by a spider."""
        src = response
        return {
            'level': "DEBUG",
            'msg': SCRAPEDMSG,
            'args': {
                'src': src,
                'item': item,
            }
        }

    def dropped(self, item, exception, response, spider):
        """Logs a message when an item is dropped while it is passing through the item pipeline."""
        return {
            'level': "WARNING",
            'msg': DROPPEDMSG,
            'args': {
                'exception': exception,
                'item': item,
            }
        }

    def item_error(self, item, exception, response, spider):
        """Logs a message when an item causes an error while it is passing
        through the item pipeline.

        .. versionadded:: 2.0
        """
        return {
            'level': "ERROR",
            'msg': ITEMERRORMSG,
            'args': {
                'item': item,
            }
        }

    def spider_error(self, failure, request, response, spider):
        """Logs an error message from a spider.

        .. versionadded:: 2.0
        """
        return {
            'level': "ERROR",
            'msg': SPIDERERRORMSG,
            'args': {
                'request': request,
                'referer': referer_str(request),
            }
        }

    def download_error(self, failure, request, spider, errmsg=None):
        """Logs a download error message from a spider (typically coming from
        the engine).

        .. versionadded:: 2.0
        """
        args = {'request': request}
        if errmsg:
            msg = DOWNLOADERRORMSG_LONG
            args['errmsg'] = errmsg
        else:
            msg = DOWNLOADERRORMSG_SHORT
        return {
            'level': "ERROR",
            'msg': msg,
            'args': args,
        }

    @classmethod
    def from_crawler(cls, crawler):
        return cls()



from aioscpy.commands import ASCommand
from aioscpy.commands.genspider import Command


class OCommand(Command):

    def short_desc(self):
        return "Generate new spider in xxx.py using pre-defined templates"

    def add_options(self, parser):
        ASCommand.add_options(self, parser)
        parser.add_argument("-l", "--list", dest="list", action="store_true",
                            help="List available templates")
        parser.add_argument("-d", "--dump", dest="dump", metavar="TEMPLATE",
                            help="Dump template to standard output")
        parser.add_argument("-t", "--template", dest="template", default="crawl",
                            help="Uses a custom template.")
        parser.add_argument("--force", dest="force", action="store_true",
                            help="If the spider already exists, overwrite it with the template")


import aioscpy
from aioscpy.commands import ASCommand


class Command(ASCommand):

    default_settings = {'LOG_ENABLED': False,
                        'SPIDER_LOADER_WARN_ONLY': True}
    requires_process = False

    def syntax(self):
        return "[-v]"

    def short_desc(self):
        return "Print aioscpy version"

    def add_options(self, parser):
        ASCommand.add_options(self, parser)
        parser.add_argument("--verbose", "-v", dest="verbose", action="store_true",
                            help="also display twisted/python/platform info (useful for bug reports)")

    def run(self, args, opts):
        # if opts.verbose:
        #     versions = aioscpy_components_versions()
        #     width = max(len(n) for (n, _) in versions)
        #     for name, version in versions:
        #         print(f"{name:<{width}} : {version}")
        # else:
        print(f"AIOSPCY {aioscpy.__version__}")



import asyncio
import traceback
import gc
import os

from time import time

from aioscpy import signals, call_grace_instance


class Slot(object):

    def __init__(self, start_requests, close_if_idle, scheduler, crawler):
        self.closing = None
        self.inprogress = set()

        self.start_requests = start_requests
        self.close_if_idle = close_if_idle
        self.scheduler = scheduler
        self.crawler = crawler
        self.heartbeat = None
        self.closing_wait = None

    def add_request(self, request):
        self.inprogress.add(request)

    def remove_request(self, request):
        self.inprogress.discard(request)

    async def close(self):
        self.closing = True


class ExecutionEngine(object):

    def __init__(self, crawler, spider_closed_callback):
        self.start_time = time()
        self.crawler = crawler
        self.settings = crawler.settings
        self.slot = None
        self.spider = None
        self.scheduler = None
        self.running = False
        self._heart_beat = None
        self._task_beat = None
        self.signals = crawler.signals
        self.logformatter = crawler.load("log_formatter")
        self.scheduler = crawler.load("scheduler")
        self.downloader = call_grace_instance('downloader', crawler)
        self.scraper = call_grace_instance("scraper", crawler)
        self.call_helper = self.di.get("tools").call_helper
        self.lock = asyncio.Lock()
        self._spider_closed_callback = spider_closed_callback

    async def start(self, spider, start_requests=None):
        self.start_time = time()
        await self.signals.send_catch_log_coroutine(signal=signals.engine_started)
        self.running = True
        await self.open_spider(spider, start_requests, close_if_idle=True)

    async def stop(self):
        self.running = False
        # await self._close_all_spiders()
        await self.signals.send_catch_log_coroutine(signal=signals.engine_stopped)

    async def start_spider_request(self, spider):
        if self.slot.start_requests and not self._needs_backout():
            try:
                request = await self.slot.start_requests.__anext__()
            except StopAsyncIteration:
                self.slot.start_requests = None
            except Exception:
                self.slot.start_requests = None
                self.logger.error('Error while obtaining start requests',
                                    exc_info=True, extra={'spider': spider})
            else:
                await self.crawl(request, spider)

    def _needs_backout(self) -> bool:
        return (
            not self.running
            or self.slot.closing
            or self.downloader.needs_backout()
            or self.scraper.slot.needs_backout()
        )

    async def _handle_downloader_output(self, result, request, spider):
        try:
            if isinstance(result, self.di.get('request')):
                await self.crawl(result, spider)
                return
            if isinstance(result, self.di.get('response')):
                result.request = request
                logkws = self.logformatter.crawled(request, result, spider)
                level, message, kwargs = self.di.get("log").logformatter_adapter(logkws)
                if logkws is not None:
                    self.logger.log(level, message, **kwargs)
                await self.signals.send_catch_log(signals.response_received,
                                                  response=result, request=request, spider=spider)
        except Exception as e:
            self.logger.error(f"enqueue_scrape: {traceback.format_exc()}")
        finally:
            self.slot.remove_request(request)
        await self.scraper.enqueue_scrape(result, request)

    async def spider_is_idle(self, spider):
        if self.scraper.slot.is_idle():
            # scraper is not idle
            return False

        if self.downloader.active:
            # downloader has pending requests
            return False

        if self.slot.start_requests is not None:
            # not all start requests are handled
            return False

        if self.slot.inprogress:
            # not all start requests are handled
            return False

        if await self.call_helper(self.slot.scheduler.has_pending_requests):
            # scheduler has pending requests
            return False

        return True

    @property
    def open_spiders(self):
        return (self.spider,) if self.spider else set()

    def has_capacity(self):
        return not bool(self.slot)

    async def crawl(self, request, spider):  # 将网址 请求加入队列
        if spider not in self.open_spiders:
            raise RuntimeError("Spider %r not opened when crawling: %s" % (spider.name, request))

        await self.signals.send_catch_log(signals.request_scheduled, request=request, spider=spider)
        if not await self.call_helper(self.slot.scheduler.enqueue_request, request):
            await self.signals.send_catch_log(signals.request_dropped, request=request, spider=spider)

    async def open_spider(self, spider, start_requests=None, close_if_idle=True):
        if not self.has_capacity():
            raise RuntimeError("No free spider slot when opening %r" % spider.name)
        self.logger.info("Spider opened({name})", **{"name": spider.name}, extra={'spider': spider})

        self.slot = Slot(start_requests, close_if_idle, self.scheduler, self.crawler)
        self.spider = spider
        await self.call_helper(self.scheduler.open, start_requests)
        await self.call_helper(self.downloader.open, spider, self)
        await self.call_helper(self.scraper.open_spider, spider)
        await self.call_helper(self.crawler.stats.open_spider, spider)
        await self.signals.send_catch_log_coroutine(signals.spider_opened, spider=spider)
        await self.start_spider_request(spider)
        if self.slot is None:
            self.logger.warning("Spider ({name}) to running not found task! please check task is be generated.",
                                **{"name": spider.name})
            return
        self._heart_beat = asyncio.create_task(self.heart_beat(5, spider, self.slot))
        self._task_beat = asyncio.create_task(self.task_beat())

    async def _close_all_spiders(self):
        dfds = [self.close_spider(s, reason='shutdown') for s in self.open_spiders]
        await asyncio.gather(*dfds)

    async def close_spider(self, spider, reason='cancelled'):
        """Close (cancel) spider and clear all its outstanding requests"""
        slot = self.slot
        if slot.closing:
            return slot.closing

        self.logger.info("Closing spider({name}) ({reason})",
                         **{'reason': reason, 'name': spider.name},
                         extra={'spider': spider})

        await slot.close()

        async def close_handler(callback, *args, errmsg='', **kwargs):
            try:
                if asyncio.iscoroutinefunction(callback):
                    await callback(*args, **kwargs)
                else:
                    callback(*args, **kwargs)
            except (Exception, BaseException) as e:
                self.logger.error(
                    errmsg + ': {exc_info}',
                    exc_info=e,
                )

        await close_handler(self.downloader.close, errmsg='Downloader close failure')

        await close_handler(self.scraper.close_spider, spider, errmsg='Scraper close failure')

        await close_handler(slot.scheduler.close, slot, errmsg='Scheduler close failure')

        await close_handler(self.signals.send_catch_log_coroutine, signal=signals.spider_closed, spider=spider,
                            reason=reason, errmsg='Error while sending spider_close signal')

        await close_handler(self.crawler.stats.close_spider, spider, reason=reason, errmsg='Stats close failure')

        self.logger.info("Spider({name}) closed ({reason})", **
                         {'reason': reason, "name": spider.name}, extra={'spider': spider})

        await self._spider_closed_callback()
        if self._heart_beat:
            self._heart_beat.cancel()
        if self._task_beat:
            self._task_beat.cancel()

    async def _spider_idle(self, spider):
        res = await self.signals.send_catch_log(signals.spider_idle, spider=spider, dont_log=self.di.get("exceptions").DontCloseSpider)
        if any(isinstance(x, self.di.get("exceptions").DontCloseSpider) for _, x in res):
            return
        if await self.spider_is_idle(spider):
            await self.close_spider(spider, reason='finished')

    async def heart_beat(self, delay, spider, slot):
        while True:
            await asyncio.sleep(delay)
            if self.running and await self.spider_is_idle(spider) and slot.close_if_idle:
                await self._spider_idle(spider)
            co = '<logstats: %(spname)s> pid: %(pid)s, transferring: %(transfer)s, queue: %(queue)s, active: %(active)s, ingress: %(ingress)s, scraper-active: %(sactive)s, scraper-queue: %(squeue)s, scraper-size: %(size)s' % {
                'spname': spider.name,
                'pid': os.getpid(),
                'transfer': len(self.downloader.slot.transferring),
                'queue': len(self.downloader.slot.queue),
                'active': len(self.downloader.active),
                'ingress': len(self.slot.inprogress),
                'sactive': len(self.scraper.slot.active),
                'squeue': len(self.scraper.slot.queue),
                'size': self.scraper.slot.active_size,
            }
            self.logger.debug(co)
            try:
                gc.collect()
            except:
                self.logger.warning(f'gc collect faild!')

    async def task_beat(self):
        local_lock = True
        while True:
            while local_lock and not self._needs_backout():
                local_lock = False
                for request in await self.slot.scheduler.async_next_request():
                    self.slot.add_request(request)
                    await self.downloader.fetch(request)
                local_lock = True
            await asyncio.sleep(1)
